# -*- coding: utf-8 -*-

"""
This module provides methods to initialize weights.
You can access them through ``brainpy.init.XXX``.
"""

from .base import *
from .generic import *
from .random_inits import *
from .regular_inits import *
from .decay_inits import *
