# -*- coding: utf-8 -*-


"""
This module provides methods to construct connectivity between neuron groups.
You can access them through ``brainpy.connect.XXX``.
"""

from .base import *
from .custom_conn import *
from .random_conn import *
from .regular_conn import *
