# -*- coding: utf-8 -*-

"""
This module aims to provide commonly used analysis methods for simulated neuronal data.
You can access them through ``brainpy.measure.XXX``.
"""

from .correlation import *
from .firings import *

